import hashlib
import os

import pytest
from django.utils.encoding import force_bytes

from wcs import fields
from wcs.admin.settings import UserFieldsFormDef
from wcs.api_access import ApiAccess
from wcs.carddef import CardDef
from wcs.qommon.http_request import HTTPRequest
from wcs.qommon.upload_storage import PicklableUpload

from ..utilities import clean_temporary_pub, create_temporary_pub, get_app
from .utils import sign_uri


@pytest.fixture
def pub(emails):
    pub = create_temporary_pub()

    req = HTTPRequest(None, {'SCRIPT_NAME': '/', 'SERVER_NAME': 'example.net'})
    pub.set_app_dir(req)
    pub.cfg['identification'] = {'methods': ['password']}
    pub.cfg['language'] = {'language': 'en'}
    pub.write_cfg()

    with open(os.path.join(pub.app_dir, 'site-options.cfg'), 'w') as fd:
        fd.write(
            '''\
[api-secrets]
coucou = 1234
'''
        )

    pub.user_class.wipe()

    return pub


def teardown_module(module):
    clean_temporary_pub()


def test_carddata_api_access(pub, local_user):
    pub.role_class.wipe()
    role = pub.role_class(name='test')
    role.store()
    local_user.roles = [role.id]
    local_user.store()

    ApiAccess.wipe()
    access = ApiAccess()
    access.name = 'test'
    access.access_identifier = 'test'
    access.access_key = '12345'
    access.store()

    CardDef.wipe()
    carddef = CardDef()
    carddef.name = 'test'
    carddef.workflow_roles = {'_viewer': role.id}
    carddef.store()

    carddef.data_class().wipe()
    carddata = carddef.data_class()()
    carddata.just_created()
    carddata.store()

    pub.custom_view_class.wipe()
    custom_view = pub.custom_view_class()
    custom_view.title = 'shared carddef custom view'
    custom_view.formdef = carddef
    custom_view.columns = {'list': [{'id': '0'}]}
    custom_view.filters = {}
    custom_view.visibility = 'any'
    custom_view.store()

    custom_view = pub.custom_view_class()
    custom_view.title = 'private carddef custom view'
    custom_view.formdef = carddef
    custom_view.columns = {'list': [{'id': '0'}]}
    custom_view.filters = {}
    custom_view.visibility = 'owner'
    custom_view.user = local_user
    custom_view.is_default = True  # check that set_default_view method is not failing with this value
    custom_view.store()

    app = get_app(pub)
    app.get(
        sign_uri(
            '/api/cards/test/shared-carddef-custom-view/%s/' % carddata.id,
            orig=access.access_identifier,
            key=access.access_key,
        ),
        status=200,
    )


@pytest.mark.parametrize('auth', ['signature', 'http-basic', 'api-admin-user'])
def test_carddata_include_params(pub, local_user, auth):
    # signature: SqlUser
    # http-basic: RestrictedApiUser
    # api-admin-user: ApiAdminUser
    pub.role_class.wipe()
    role = pub.role_class(name='test')
    role.id = '123'
    role.store()
    local_user.roles = [role.id]
    local_user.store()

    CardDef.wipe()
    carddef = CardDef()
    carddef.name = 'test'
    carddef.workflow_roles = {'_viewer': role.id}
    carddef.fields = [
        fields.StringField(id='1', label='foobar', varname='foobar'),
    ]
    carddef.store()

    carddef.data_class().wipe()
    carddata = carddef.data_class()()
    carddata.data = {'1': 'FOO BAR'}
    carddata.user_id = local_user.id
    carddata.just_created()
    upload = PicklableUpload('test.txt', 'text/plain', 'ascii')
    upload.receive([b'test'])
    upload2 = PicklableUpload('test.txt', 'text/plain', 'ascii')
    upload2.receive([b'test'])
    carddata.workflow_data = {'blah': upload, 'blah2': upload2, 'xxx': 23}
    carddata.store()

    ApiAccess.wipe()
    access = ApiAccess()
    access.name = 'test'
    access.access_identifier = 'test'
    access.access_key = '12345'
    access.store()

    app = get_app(pub)

    if auth == 'http-basic':
        # there's not "defaults to admin" permissions in case of basic authentication.
        access.roles = [role]
        access.store()

        def get_url(url, **kwargs):
            app.set_authorization(('Basic', ('test', '12345')))
            return app.get(url, **kwargs)

    else:

        def get_url(url, **kwargs):
            return app.get(
                sign_uri(
                    url,
                    user=local_user if auth == 'signature' else None,
                    orig=access.access_identifier,
                    key=access.access_key,
                ),
                **kwargs,
            )

    resp = get_url('/api/cards/test/list?include-fields=on')
    assert 'fields' in resp.json['data'][0]
    resp = get_url('/api/cards/test/list?include-evolution=on')
    assert 'evolution' in resp.json['data'][0]
    resp = get_url('/api/cards/test/list?include-roles=on')
    assert 'roles' in resp.json['data'][0]
    resp = get_url('/api/cards/test/list?include-submission=on')
    assert 'submission' in resp.json['data'][0]
    resp = get_url('/api/cards/test/list?include-workflow=on')
    assert 'workflow' in resp.json['data'][0]
    assert 'data' not in resp.json['data'][0]['workflow']
    resp = get_url('/api/cards/test/list?include-workflow-data=on')
    assert 'workflow' in resp.json['data'][0]
    assert 'data' in resp.json['data'][0]['workflow']

    resp = get_url('/api/cards/test/%s/?include-fields=off' % carddata.id)
    assert 'fields' not in resp.json
    resp = get_url('/api/cards/test/%s/?include-evolution=off' % carddata.id)
    assert 'evolution' not in resp.json
    resp = get_url('/api/cards/test/%s/?include-roles=off' % carddata.id)
    assert 'roles' not in resp.json
    resp = get_url('/api/cards/test/%s/?include-submission=off' % carddata.id)
    assert 'submission' not in resp.json
    resp = get_url('/api/cards/test/%s/?include-workflow=off' % carddata.id)
    assert 'workflow' in resp.json
    assert 'data' in resp.json['workflow']
    resp = get_url('/api/cards/test/%s/?include-workflow-data=off' % carddata.id)
    assert 'workflow' in resp.json
    assert 'data' not in resp.json['workflow']
    resp = get_url('/api/cards/test/%s/?include-workflow=off&include-workflow-data=off' % carddata.id)
    assert 'workflow' not in resp.json


def test_carddata_user_fields(pub, local_user):
    pub.role_class.wipe()
    role = pub.role_class(name='test')
    role.store()

    local_user.roles = [role.id]
    local_user.store()

    CardDef.wipe()
    carddef = CardDef()
    carddef.name = 'test'
    carddef.workflow_roles = {'_viewer': role.id}
    carddef.store()

    carddata = carddef.data_class()()
    carddata.just_created()
    carddata.store()

    resp = get_app(pub).get(sign_uri('/api/cards/test/%s/' % carddata.id, user=local_user), status=200)
    assert 'user' not in resp.json
    resp = get_app(pub).get(sign_uri('/api/cards/test/list', user=local_user))
    assert 'user' not in resp.json['data'][0]
    resp = get_app(pub).get(sign_uri('/api/cards/test/list?full=on', user=local_user))
    assert 'user' not in resp.json['data'][0]

    carddata.user_id = local_user.id
    carddata.store()
    resp = get_app(pub).get(sign_uri('/api/cards/test/%s/' % carddata.id, user=local_user), status=200)
    assert resp.json['user'] == {
        'id': local_user.id,
        'NameID': ['0123456789'],
        'name': 'Jean Darmette',
        'email': 'jean.darmette@triffouilis.fr',
    }
    resp = get_app(pub).get(sign_uri('/api/cards/test/list', user=local_user))
    assert 'user' not in resp.json['data'][0]
    resp = get_app(pub).get(sign_uri('/api/cards/test/list?full=on', user=local_user))
    assert resp.json['data'][0]['user'] == {
        'id': local_user.id,
        'NameID': ['0123456789'],
        'name': 'Jean Darmette',
        'email': 'jean.darmette@triffouilis.fr',
    }

    formdef = UserFieldsFormDef(pub)
    formdef.fields = [
        fields.StringField(id='3', label='test', type='string', varname='var1'),
        fields.StringField(id='9', label='noop', type='string', varname='var2'),
        fields.StringField(id='42', label='no varname', type='string'),
    ]
    formdef.store()

    resp = get_app(pub).get(sign_uri('/api/cards/test/%s/' % carddata.id, user=local_user), status=200)
    assert resp.json['user'] == {
        'id': local_user.id,
        'NameID': ['0123456789'],
        'name': 'Jean Darmette',
        'email': 'jean.darmette@triffouilis.fr',
        'var1': None,
        'var2': None,
    }
    resp = get_app(pub).get(sign_uri('/api/cards/test/list', user=local_user))
    assert 'user' not in resp.json['data'][0]
    resp = get_app(pub).get(sign_uri('/api/cards/test/list?full=on', user=local_user))
    assert resp.json['data'][0]['user'] == {
        'id': local_user.id,
        'NameID': ['0123456789'],
        'name': 'Jean Darmette',
        'email': 'jean.darmette@triffouilis.fr',
        'var1': None,
        'var2': None,
    }

    local_user.form_data = {'3': 'toto', '9': 'nono'}
    local_user.store()
    resp = get_app(pub).get(sign_uri('/api/cards/test/%s/' % carddata.id, user=local_user), status=200)
    assert resp.json['user'] == {
        'id': local_user.id,
        'NameID': ['0123456789'],
        'name': 'Jean Darmette',
        'email': 'jean.darmette@triffouilis.fr',
        'var1': 'toto',
        'var2': 'nono',
    }
    resp = get_app(pub).get(sign_uri('/api/cards/test/list', user=local_user))
    assert 'user' not in resp.json['data'][0]
    resp = get_app(pub).get(sign_uri('/api/cards/test/list?full=on', user=local_user))
    assert resp.json['data'][0]['user'] == {
        'id': local_user.id,
        'NameID': ['0123456789'],
        'name': 'Jean Darmette',
        'email': 'jean.darmette@triffouilis.fr',
        'var1': 'toto',
        'var2': 'nono',
    }


def test_card_get_file(pub):
    pub.role_class.wipe()

    CardDef.wipe()
    carddef = CardDef()
    carddef.name = 'test'
    carddef.fields = [
        fields.StringField(id='0', label='foobar', varname='foobar'),
        fields.FileField(id='3', label='foobar4', varname='file'),
    ]
    carddef.store()

    upload = PicklableUpload('test.txt', 'text/plain', 'ascii')
    upload.receive([b'file content'])

    carddef.data_class().wipe()
    formdata = carddef.data_class()()
    formdata.data = {'0': 'blah', '3': upload}
    formdata.just_created()
    formdata.store()

    resp = get_app(pub).get(sign_uri('/api/cards/test/%s/' % formdata.id))
    file_url = resp.json['fields']['file']['url']
    assert get_app(pub).get(file_url, status=403)

    resp = get_app(pub).get(sign_uri(file_url), status=200)
    assert resp.text == 'file content'

    upload = PicklableUpload('test.jpeg', 'image/jpeg')
    with open(os.path.join(os.path.dirname(__file__), '../image-with-gps-data.jpeg'), 'rb') as fd:
        upload.receive([fd.read()])
    formdata.data = {'0': 'blah', '3': upload}
    formdata.store()

    resp = get_app(pub).get(sign_uri('/api/cards/test/%s/' % formdata.id))
    file_url = resp.json['fields']['file']['url']
    assert upload.can_thumbnail() is True
    thumbs_dir = os.path.join(pub.app_dir, 'thumbs')
    thumb_filepath = os.path.join(
        thumbs_dir, hashlib.sha256(force_bytes(upload.get_fs_filename())).hexdigest()
    )
    assert os.path.exists(thumbs_dir) is False
    assert os.path.exists(thumb_filepath) is False
    get_app(pub).get(sign_uri(file_url + '&thumbnail=1'), status=200)
    assert os.path.exists(thumbs_dir) is True
    assert os.path.exists(thumb_filepath) is True
    # again, thumbs_dir already exists
    get_app(pub).get(sign_uri(file_url + '&thumbnail=1'), status=200)


def test_api_list_formdata_phone_order_by_rank(pub):
    pub.role_class.wipe()
    role = pub.role_class(name='test')
    role.store()

    ApiAccess.wipe()
    access = ApiAccess()
    access.name = 'test'
    access.access_identifier = 'test'
    access.access_key = '12345'
    access.store()

    CardDef.wipe()
    carddef = CardDef()
    carddef.name = 'test'
    carddef.workflow_roles = {'_receiver': role.id}
    carddef.fields = [
        fields.StringField(id='0', label='a', type='string', display_locations=['listings']),
        fields.StringField(id='1', label='b', type='string'),
    ]
    carddef.store()

    data_class = carddef.data_class()
    data_class.wipe()

    # 1st carddata, with phone number
    carddata1 = data_class()
    carddata1.data = {'1': '+33623456789'}
    carddata1.just_created()
    carddata1.jump_status('new')
    carddata1.store()

    # 2nd carddata, with no value
    carddata2 = data_class()
    carddata2.data = {}
    carddata2.just_created()
    carddata2.jump_status('new')
    carddata2.store()

    # check fts
    resp = get_app(pub).get(
        sign_uri(
            '/api/cards/test/list?full=on&q=0623456789', orig=access.access_identifier, key=access.access_key
        )
    )
    assert len(resp.json['data']) == 1
    assert [int(x['id']) for x in resp.json['data']] == [carddata1.id]


def test_carddata_list_skip_evolutions(pub, sql_queries):
    pub.role_class.wipe()
    role = pub.role_class(name='test')
    role.id = '123'
    role.store()

    CardDef.wipe()
    carddef = CardDef()
    carddef.name = 'test'
    carddef.workflow_roles = {'_viewer': role.id}
    carddef.fields = [
        fields.StringField(id='1', label='foobar', varname='foobar'),
    ]
    carddef.store()

    carddef.data_class().wipe()

    for i in range(10):
        carddata = carddef.data_class()()
        carddata.data = {'1': 'FOO BAR %s' % i}
        carddata.just_created()
        carddata.store()

    ApiAccess.wipe()
    access = ApiAccess()
    access.name = 'test'
    access.access_identifier = 'test'
    access.access_key = '12345'
    access.roles = [role]
    access.store()

    app = get_app(pub)

    def get_url(url, **kwargs):
        app.set_authorization(('Basic', ('test', '12345')))
        return app.get(url, **kwargs)

    sql_queries.clear()
    get_url('/api/cards/test/list')
    assert not [x for x in sql_queries if '%s_evolution' % carddef.table_name in x]
    sql_queries.clear()
    get_url('/api/cards/test/list?include-workflow=on')
    assert [x for x in sql_queries if '%s_evolution' % carddef.table_name in x]
    sql_queries.clear()
    get_url('/api/cards/test/list?include-evolution=on')
    assert [x for x in sql_queries if '%s_evolution' % carddef.table_name in x]
    sql_queries.clear()
